import { invoke } from '@tauri-apps/api/core';
import dayjs from 'dayjs';
import { atom, useAtomValue, useSetAtom } from 'jotai';
import { atomWithRefresh } from 'jotai/utils';
import { get, reduce } from 'lodash-es';
import { useCallback } from 'react';
import { v4 } from 'uuid';
import { NotificationType, ToastDuration, useNotification } from '../components/Notifications';

export enum FrequencyShifting {
  /**
   * Change frequency undergoes a linear transformation from previous pulse to current one.
   */
  Linear,
  /**
   * Change frequency undergoes a quadratic transformation from previous pulse to current one.
   */
  Quadratic,
  /**
   * Change frequency undergoes a cubic transformation from previous pulse to current one.
   */
  Cubic,
  /**
   * Change frequency with quick fade in and fade out.
   */
  Ease,
  /**
   * Change frequency with spiking within range from previous pulse and current one.
   */
  Pulsating,
  /**
   * Based on frequency of previous pulse, take the twice of frequency of current pulse as the peak frequency. Follow the maximum frequency limitation.
   */
  Spiking,
  /**
   * Randomize frequency within range from previous frequency and current one.
   */
  Randomize,
  /**
   * Randomize frequency within minium and maximum frequency.
   */
  Maniac,
  /**
   * Synchronize changes of frequency with pulse width changes.
   */
  Synchronized,
}

export interface ControlPoint {
  x: number;
  y: number;
}

export class Pulse {
  order: number;
  id: number;
  offset: number;
  width: number;
  maniac: boolean;
  frequency: number;
  frequencyShifting: FrequencyShifting;
  controlPoint1: ControlPoint;
  controlPoint2: ControlPoint;

  constructor(order: number, width: number, frequency: number) {
    this.id = v4();
    this.order = order;
    this.offset = 0;
    this.width = width;
    this.maniac = false;
    this.frequency = frequency;
    this.frequencyShifting = FrequencyShifting.Linear;
    this.controlPoint1 = { x: 0, y: 0 };
    this.controlPoint2 = { x: 0, y: 0 };
  }

  equals(other: Pulse): boolean {
    return this.id === other.id;
  }
}

export class Pattern {
  id: string;
  name: string;
  createdAt: number;
  lastModifiedAt: number | null;
  smoothRepeat: boolean;
  pulses: Pulse[];

  constructor() {
    this.id = v4();
    this.name = '';
    this.createdAt = dayjs().valueOf();
    this.lastModifiedAt = null;
    this.smoothRepeat = true;
    this.pulses = [];
  }

  movePulseUp(pulseId: string, step: number) {
    const index = this.pulses.findIndex((pulse) => pulse.id === pulseId);
    if (index === -1 || index - step < 0) return;

    const targetIndex = index - step;
    const targetPulse = this.pulses[targetIndex];
    const currentPulse = this.pulses[index];

    // Swap the pulses
    this.pulses[targetIndex] = currentPulse;
    this.pulses[index] = targetPulse;

    // Swap their order
    const tempOrder = currentPulse.order;
    currentPulse.order = targetPulse.order;
    targetPulse.order = tempOrder;

    // Sort pulses by order
    this.pulses.sort((a, b) => a.order - b.order);
  }

  movePulseDown(pulseId: string, step: number) {
    const index = this.pulses.findIndex((pulse) => pulse.id === pulseId);
    if (index === -1 || index + step >= this.pulses.length) return;

    const targetIndex = index + step;
    const targetPulse = this.pulses[targetIndex];
    const currentPulse = this.pulses[index];

    // Swap the pulses
    this.pulses[targetIndex] = currentPulse;
    this.pulses[index] = targetPulse;

    // Swap their order
    const tempOrder = currentPulse.order;
    currentPulse.order = targetPulse.order;
    targetPulse.order = tempOrder;

    // Sort pulses by order
    this.pulses.sort((a, b) => a.order - b.order);
  }

  addPulse(): Pulse {
    const maxOrder = reduce(this.pulses, (former, pulse) => Math.max(former, pulse.order), 0);
    const newPulse = new Pulse(
      maxOrder + 1,
      this.smoothRepeat ? get(this.pulses, '[0].width', 0) : get(this.pulses, '[-1].width', 0),
      this.smoothRepeat
        ? get(this.pulses, '[0].frequency', 1)
        : get(this.pulses, '[-1].frequency', 1),
    );
    this.pulses.push(newPulse);
    return newPulse;
  }

  updatePulse(pulseId: string, pulse: Pulse) {
    const index = this.pulses.findIndex((p) => p.id === pulseId);
    if (index !== -1) {
      const { id, order, ...rest } = pulse;
      this.pulses[index] = { ...this.pulses[index], ...rest };
    }
  }

  deletePulse(pulseId: string) {
    this.pulses = this.pulses.filter((pulse) => pulse.id !== pulseId);
    this.pulses.sort((a, b) => a.order - b.order);
    this.pulses.forEach((pulse, index) => {
      pulse.order = index + 1;
    });
  }
}

export function totalDuration(pattern: Pattern): number {
  return reduce(
    pattern.pulses,
    (former, pulse) => former + pulse.offset,
    pattern.smoothRepeat && pattern.pulses.length > 1 ? 100 : 0,
  );
}

export const SearchKeywordAtom = atom<string | null>(null);
export const PatternsAtom = atomWithRefresh(async (get) => {
  try {
    const keyword = get(SearchKeywordAtom);
    const patterns = await invoke<Pattern[]>('list_patterns', { keyword });
    return patterns;
  } catch (e) {
    console.error('[retrieving pattern list]', e);
  }
  return [];
});
export const SelectedPatternIdAtom = atom<string | null>(null);
export const CurrentPatternAtom = atomWithRefresh<Pattern | null>(async (get) => {
  try {
    const patternId = get(SelectedPatternIdAtom);
    if (patternId === null) {
      return null;
    }
    const pattern = await invoke('get_pattern', { patternId });
    return pattern;
  } catch (e) {
    console.error('[retrieving pattern]', e);
  }
  return null;
});
export const PulsesInCurrentPatternAtom = atomWithRefresh(
  (get) => get(CurrentPatternAtom)?.pulses ?? [],
);
export const CurrentPatternDuration = atom((get) => {
  const currentPattern = get(CurrentPatternAtom);
  if (!currentPattern) return 0;
  return totalDuration(currentPattern);
});
export const SelectedPulseIdAtom = atom<string | null>(null);
export const SelectedPulseAtom = atom<Pulse | null>((get) => {
  const pulses = get(PulsesInCurrentPatternAtom);
  const selectedPulseId = get(SelectedPulseIdAtom);
  return pulses.find((pulse) => pulse.id === selectedPulseId) ?? null;
});

export function useSavePattern() {
  const refreshPatterns = useSetAtom(PatternsAtom);
  const selectedPatternId = useAtomValue(SelectedPatternIdAtom);
  const refreshSelectedPattern = useSetAtom(CurrentPatternAtom);
  const { showToast } = useNotification();

  const savePattern = useCallback(
    async (pattern: Pattern) => {
      try {
        await invoke('save_pattern', { pattern });
        refreshPatterns();
        if (pattern.id === selectedPatternId) {
          refreshSelectedPattern();
        }
        return true;
      } catch (error) {
        console.error('[save pattern]', error);
        showToast(
          NotificationType.ERROR,
          'Failed to save pattern. Please try again.',
          'material-symbols-light:error-outline',
          ToastDuration.MEDIUM,
        );
      }
      return false;
    },
    [selectedPatternId],
  );

  return savePattern;
}
